@@ -26,10 +26,12 @@
 #include <stdbool.h>
 #include <assert.h>
 
+#include "init.h"
 #include "ica_api.h"
 #include "icastats.h"
 #include "fips.h"
 #include "s390_rsa.h"
+#include "s390_ecc.h"
 #include "s390_crypto.h"
 #include "s390_sha.h"
 #include "s390_prng.h"
@@ -47,6 +49,16 @@
 
 #define MAX_VERSION_LENGTH 16
 
+int ica_fallbacks_enabled = 1;
+
+void ica_set_fallback_mode(int fallback_mode)
+{
+	if (fallback_mode)
+		ica_fallbacks_enabled = 1;
+	else
+		ica_fallbacks_enabled = 0;
+}
+
 static unsigned int check_des_parms(unsigned int mode,
 				    unsigned long data_length,
 				    const unsigned char *in_data,
@@ -218,15 +230,31 @@ static unsigned int check_cmac_parms(unsigned int block_size,
 }
 
 static unsigned int check_gcm_parms(unsigned long text_length,
-				    const unsigned char *aad,
 				    unsigned long aad_length,
 				    const unsigned char *tag, unsigned int tag_length,
 				    unsigned int iv_length)
 {
+#ifdef __s390x__
+	/*
+	 * The following comparisions are alaways false on s390 targets
+	 * due to limited range of data type.
+	 */
 	if ((text_length > S390_GCM_MAX_TEXT_LENGTH) ||
-	    (aad_length  > S390_GCM_MAX_AAD_LENGTH) ||
-	    (iv_length   > S390_GCM_MAX_IV_LENGTH) ||
-	    (iv_length == 0))
+	    (aad_length > S390_GCM_MAX_AAD_LENGTH))
+		return EINVAL;
+#else
+	(void)text_length;	/* supporess unused param warning */
+	(void)aad_length;
+#endif
+	/*
+	 * The following check must be done but is commented out because
+	 * comparison is always false due to limited range of data type.
+	 *
+	 * if (iv_length   > S390_GCM_MAX_IV_LENGTH)
+	 *	return EINVAL;
+	 */
+
+	if (iv_length == 0)
 		return EINVAL;
 
 	if (tag == NULL)
@@ -249,7 +277,6 @@ static unsigned int check_gcm_parms(unsigned long text_length,
 }
 
 static unsigned int check_ccm_parms(unsigned long payload_length,
-				    const unsigned char *assoc_data,
 				    unsigned long assoc_data_length,
 				    const unsigned char *mac,
 				    unsigned int mac_length,
@@ -895,13 +922,15 @@ unsigned int ica_rsa_mod_expo(ica_adapter_handle_t adapter_handle,
 
 	hardware = ALGO_SW;
 	if (adapter_handle == DRIVER_NOT_LOADED)
-		rc = rsa_mod_expo_sw(&rb);
+		rc = ica_fallbacks_enabled ?
+			rsa_mod_expo_sw(&rb) : ENODEV;
 	else {
 		rc = ioctl(adapter_handle, ICARSAMODEXPO, &rb);
 		if (!rc)
 			hardware = ALGO_HW;
 		else
-			rc = rsa_mod_expo_sw(&rb);
+			rc = ica_fallbacks_enabled ?
+				rsa_mod_expo_sw(&rb) : ENODEV;
 	}
 	if (rc == 0)
 		stats_increment(ICA_STATS_RSA_ME, hardware, ENCRYPT);
@@ -1002,13 +1031,15 @@ unsigned int ica_rsa_crt(ica_adapter_handle_t adapter_handle,
 
 	hardware = ALGO_SW;
 	if (adapter_handle == DRIVER_NOT_LOADED)
-		rc = rsa_crt_sw(&rb);
+		rc = ica_fallbacks_enabled ?
+			rsa_crt_sw(&rb) : ENODEV;
 	else {
 		rc = ioctl(adapter_handle, ICARSACRT, &rb);
 		if(!rc)
 			hardware = ALGO_HW;
 		else
-			rc = rsa_crt_sw(&rb);
+			rc = ica_fallbacks_enabled ?
+				rsa_crt_sw(&rb) : ENODEV;
 	}
 	if (rc == 0)
 		stats_increment(ICA_STATS_RSA_CRT, hardware, ENCRYPT);
@@ -1016,6 +1047,286 @@ unsigned int ica_rsa_crt(ica_adapter_handle_t adapter_handle,
 	return rc;
 }
 
+/*******************************************************************************
+ *
+ *                          Begin of ECC API
+ */
+
+ICA_EC_KEY* ica_ec_key_new(unsigned int nid, unsigned int *privlen)
+{
+	ICA_EC_KEY *key;
+	int len;
+
+#ifdef ICA_FIPS
+	if (fips >> 1)
+		return NULL;
+#endif /* ICA_FIPS */
+
+	/* check for obvious errors in parms */
+	if (!curve_supported(nid))
+		return NULL;
+
+	if ((key = malloc(sizeof(ICA_EC_KEY))) == NULL)
+		return NULL;
+
+	/* allocate clear memory for the 3 key parts */
+	len = privlen_from_nid(nid);
+	key->X = calloc(1, 3*len);
+	if (!key->X)
+		return NULL;
+
+	key->nid = nid;
+	key->Y = key->X + len;
+	key->D = key->Y + len;
+
+	*privlen = len;
+
+	return key;
+}
+
+int ica_ec_key_init(const unsigned char *X, const unsigned char *Y,
+		const unsigned char *D, ICA_EC_KEY *key)
+{
+	unsigned int privlen = privlen_from_nid(key->nid);
+
+#ifdef ICA_FIPS
+	if (fips >> 1)
+		return EACCES;
+#endif /* ICA_FIPS */
+
+	/* check for obvious errors in parms */
+	if (key == NULL)
+		return EINVAL;
+
+	if ((X == NULL && Y != NULL) || (X != NULL && Y == NULL))
+		return EINVAL;
+
+	if (X != NULL && Y != NULL) {
+		memcpy(key->X, X, privlen);
+		memcpy(key->Y, Y, privlen);
+	}
+
+	if (D != NULL)
+		memcpy(key->D, D, privlen);
+
+	return 0;
+}
+
+int ica_ec_key_generate(ica_adapter_handle_t adapter_handle, ICA_EC_KEY *key)
+{
+	int hardware, rc;
+	unsigned int icapath = 0;
+
+#ifdef ICA_FIPS
+	if (fips >> 1)
+		return EACCES;
+#endif /* ICA_FIPS */
+
+	/* check for obvious errors in parms */
+	if (key == NULL)
+		return EINVAL;
+
+	icapath = getenv_icapath();
+	switch (icapath) {
+	case 1: /* hw only */
+		hardware = ALGO_HW;
+		rc = eckeygen_hw(adapter_handle, key);
+		break;
+	case 2: /* sw only */
+		hardware = ALGO_SW;
+		rc = eckeygen_sw(key);
+		break;
+	default: /* hw with sw fallback (default) */
+		hardware = ALGO_SW;
+		rc = eckeygen_hw(adapter_handle, key);
+		if (rc == 0)
+			hardware = ALGO_HW;
+		else
+			rc = ica_fallbacks_enabled ?
+				eckeygen_sw(key) : ENODEV;
+	}
+
+	if (rc == 0)
+		stats_increment(ICA_STATS_ECKGEN, hardware, ENCRYPT);
+
+	return rc;
+}
+
+int ica_ecdh_derive_secret(ica_adapter_handle_t adapter_handle,
+		const ICA_EC_KEY *privkey_A, const ICA_EC_KEY *pubkey_B,
+		unsigned char *z, unsigned int z_length)
+{
+	int hardware, rc;
+	unsigned int privlen = privlen_from_nid(privkey_A->nid);
+	unsigned int icapath = 0;
+
+#ifdef ICA_FIPS
+	if (fips >> 1)
+		return EACCES;
+#endif /* ICA_FIPS */
+
+	/* check for obvious errors in parms */
+	if (z == NULL || z_length < privlen || privkey_A->nid != pubkey_B->nid)
+		return EINVAL;
+
+	icapath = getenv_icapath();
+	switch (icapath) {
+	case 1: /* hw only */
+		hardware = ALGO_HW;
+		rc = ecdh_hw(adapter_handle, privkey_A, pubkey_B, z);
+		break;
+	case 2: /* sw only */
+		hardware = ALGO_SW;
+		rc = ecdh_sw(privkey_A, pubkey_B, z);
+		break;
+	default: /* hw with sw fallback (default) */
+		hardware = ALGO_SW;
+		rc = ecdh_hw(adapter_handle, privkey_A, pubkey_B, z);
+		if (rc == 0)
+			hardware = ALGO_HW;
+		else
+			rc = ica_fallbacks_enabled ?
+				ecdh_sw(privkey_A, pubkey_B, z) : ENODEV;
+	}
+
+	if (rc == 0)
+		stats_increment(ICA_STATS_ECDH, hardware, ENCRYPT);
+
+	return rc;
+}
+
+int ica_ecdsa_sign(ica_adapter_handle_t adapter_handle,
+		const ICA_EC_KEY *privkey, const unsigned char *hash, unsigned int hash_length,
+		unsigned char *signature, unsigned int signature_length)
+{
+	int hardware, rc;
+	unsigned int privlen = privlen_from_nid(privkey->nid);
+	unsigned int icapath = 0;
+
+#ifdef ICA_FIPS
+	if (fips >> 1)
+		return EACCES;
+#endif /* ICA_FIPS */
+
+	/* check for obvious errors in parms */
+	if (privkey == NULL || hash == NULL || !hash_length_valid(hash_length) ||
+		signature == NULL || signature_length < 2*privlen)
+		return EINVAL;
+
+	icapath = getenv_icapath();
+	switch (icapath) {
+	case 1: /* hw only */
+		hardware = ALGO_HW;
+		rc = ecdsa_sign_hw(adapter_handle, privkey, hash, hash_length, signature);
+		break;
+	case 2: /* sw only */
+		hardware = ALGO_SW;
+		rc = ecdsa_sign_sw(privkey, hash, hash_length, signature);
+		break;
+	default: /* hw with sw fallback (default) */
+		hardware = ALGO_SW;
+		rc = ecdsa_sign_hw(adapter_handle, privkey, hash, hash_length, signature);
+		if (rc == 0)
+			hardware = ALGO_HW;
+		else
+			rc = ica_fallbacks_enabled ?
+				ecdsa_sign_sw(privkey, hash, hash_length, signature) : ENODEV;
+	}
+
+	if (rc == 0)
+		stats_increment(ICA_STATS_ECDSA_SIGN, hardware, ENCRYPT);
+
+	return rc;
+}
+
+int ica_ecdsa_verify(ica_adapter_handle_t adapter_handle,
+		const ICA_EC_KEY *pubkey, const unsigned char *hash, unsigned int hash_length,
+		const unsigned char *signature, unsigned int signature_length)
+{
+	int hardware, rc;
+	unsigned int privlen = privlen_from_nid(pubkey->nid);
+	unsigned int icapath = 0;
+
+#ifdef ICA_FIPS
+	if (fips >> 1)
+		return EACCES;
+#endif /* ICA_FIPS */
+
+	/* check for obvious errors in parms */
+	if (pubkey == NULL || hash == NULL || !hash_length_valid(hash_length) ||
+		signature == NULL || signature_length < 2*privlen)
+		return EINVAL;
+
+	icapath = getenv_icapath();
+	switch (icapath) {
+	case 1: /* hw only */
+		hardware = ALGO_HW;
+		rc = ecdsa_verify_hw(adapter_handle, pubkey, hash, hash_length, signature);
+		break;
+	case 2: /* sw only */
+		hardware = ALGO_SW;
+		rc = ecdsa_verify_sw(pubkey, hash, hash_length, signature);
+		break;
+	default: /* hw with sw fallback (default) */
+		hardware = ALGO_SW;
+		rc = ecdsa_verify_hw(adapter_handle, pubkey, hash, hash_length, signature);
+		if (rc == 0) {
+			hardware = ALGO_HW;
+		} else if (rc != EFAULT) {
+			rc = ica_fallbacks_enabled ?
+			     ecdsa_verify_sw(pubkey, hash, hash_length,
+					     signature) : ENODEV;
+		}
+	}
+
+	if (rc == 0)
+		stats_increment(ICA_STATS_ECDSA_VERIFY, hardware, ENCRYPT);
+
+	return rc;
+}
+
+int ica_ec_key_get_public_key(const ICA_EC_KEY *key, unsigned char *q, unsigned int *q_len)
+{
+	if (!key || !(key->X) || privlen_from_nid(key->nid) < 0)
+		return EINVAL;
+
+	memcpy(q, key->X, 2*privlen_from_nid(key->nid));
+	*q_len = 2*privlen_from_nid(key->nid);
+
+	return 0;
+}
+
+int ica_ec_key_get_private_key(const ICA_EC_KEY *key, unsigned char *d, unsigned int *d_len)
+{
+	if (!key || !(key->D) || privlen_from_nid(key->nid) < 0)
+		return EINVAL;
+
+	memcpy(d, key->D, privlen_from_nid(key->nid));
+	*d_len = privlen_from_nid(key->nid);
+
+	return 0;
+}
+
+void ica_ec_key_free(ICA_EC_KEY *key)
+{
+	if (!key)
+		return;
+
+	if (key->X) {
+		/* free 1 block of memory for X, Y, and D */
+		OPENSSL_cleanse((void *)key->X, 3*privlen_from_nid(key->nid));
+		free(key->X);
+	}
+
+	OPENSSL_cleanse((void *)key, sizeof(ICA_EC_KEY));
+	free(key);
+}
+
+/*
+ *                             End of ECC API
+ *
+ ******************************************************************************/
+
 unsigned int ica_des_encrypt(unsigned int mode,
 			     unsigned int data_length,
 			     unsigned char *input_data,
@@ -1427,7 +1738,7 @@ unsigned int ica_des_cmac_last(const unsigned char *message, unsigned long messa
 			       DES_BLOCK_SIZE, key, mac_length, tmp_mac, iv);
 		if (rc)
 			return rc;
-		if (memcmp(tmp_mac, mac, mac_length))
+		if (CRYPTO_memcmp(tmp_mac, mac, mac_length))
 			return EFAULT;
 		else
 			stats_increment(ICA_STATS_DES_CMAC, ALGO_HW, direction);
@@ -1671,7 +1982,7 @@ unsigned int ica_3des_cmac_last(const unsigned char *message, unsigned long mess
 			       3*DES_BLOCK_SIZE, key, mac_length, tmp_mac, iv);
 		if (rc)
 			return rc;
-		if (memcmp(tmp_mac, mac, mac_length))
+		if (CRYPTO_memcmp(tmp_mac, mac, mac_length))
 			return EFAULT;
 		else
 			stats_increment(ICA_STATS_3DES_CMAC, ALGO_HW, direction);
@@ -1739,7 +2050,7 @@ unsigned int ica_aes_cbc_cs(const unsigned char *in_data, unsigned char *out_dat
 
 	function_code = aes_directed_fc(key_length, direction);
 	return s390_aes_cbccs(function_code, in_data, out_data, data_length,
-			      key, key_length, iv, variant);
+			      key, iv, variant);
 }
 
 unsigned int ica_aes_cfb(const unsigned char *in_data, unsigned char *out_data,
@@ -1960,7 +2271,7 @@ unsigned int ica_aes_cmac_last(const unsigned char *message, unsigned long messa
 			       key_length, key, mac_length, tmp_mac, iv);
 		if (rc)
 			return rc;
-		if (memcmp(tmp_mac, mac, mac_length))
+		if (CRYPTO_memcmp(tmp_mac, mac, mac_length))
 			return EFAULT;
 		else
 			stats_increment(ICA_STATS_AES_CMAC, ALGO_HW, direction);
@@ -1989,8 +2300,7 @@ unsigned int ica_aes_ccm(unsigned char *payload, unsigned long payload_length,
 	if (check_aes_parms(MODE_CCM, payload_length, payload, nonce, key_length,
 			    key, ciphertext_n_mac))
 		return EINVAL;
-	if (check_ccm_parms(payload_length,
-			    assoc_data, assoc_data_length,
+	if (check_ccm_parms(payload_length, assoc_data_length,
 			    ciphertext_n_mac + payload_length, mac_length,
 			    nonce_length))
 		return EINVAL;
@@ -2012,7 +2322,7 @@ unsigned int ica_aes_ccm(unsigned char *payload, unsigned long payload_length,
 
 	if (direction == ICA_DECRYPT) {
 		/* verify */
-		if (memcmp((unsigned char *)(ciphertext_n_mac + payload_length),
+		if (CRYPTO_memcmp((unsigned char *)(ciphertext_n_mac + payload_length),
 			   tmp_mac, mac_length))
 			return EFAULT;
 	}
@@ -2047,7 +2357,7 @@ unsigned int ica_aes_gcm(unsigned char *plaintext, unsigned long plaintext_lengt
 				    iv, key_length, key, (unsigned char *)1))
 			return EINVAL;
 	}
-	if (check_gcm_parms(plaintext_length, aad, aad_length, tag, tag_length, iv_length))
+	if (check_gcm_parms(plaintext_length, aad_length, tag, tag_length, iv_length))
 		return EINVAL;
 
 	memset(tmp_tag, 0, sizeof(tmp_tag));
@@ -2076,7 +2386,7 @@ unsigned int ica_aes_gcm(unsigned char *plaintext, unsigned long plaintext_lengt
 		if (rc)
 			return rc;
 
-		if (memcmp(tmp_tag, tag, tag_length))
+		if (CRYPTO_memcmp(tmp_tag, tag, tag_length))
 			return EFAULT;
 	}
 	return 0;
@@ -2131,21 +2441,21 @@ unsigned int ica_aes_gcm_intermediate(unsigned char *plaintext,
 				    cb, key_length, key, (unsigned char *)1))
 			return EINVAL;
 	}
-	if (check_gcm_parms(plaintext_length, aad, aad_length, tag, tag_length,
-				iv_length_dummy))
+	if (check_gcm_parms(plaintext_length, aad_length, tag, tag_length,
+			    iv_length_dummy))
 		return EINVAL;
 
 	function_code = aes_directed_fc(key_length, direction);
 	if (direction) {
 		/* encrypt & generate */
 		rc = s390_gcm_intermediate(function_code, plaintext, plaintext_length,
-		    ciphertext, cb, aad, aad_length, tag, tag_length, key, subkey);
+		    ciphertext, cb, aad, aad_length, tag, key, subkey);
 		if (rc)
 			return rc;
 	} else {
 		/* decrypt & verify */
 		rc = s390_gcm_intermediate(function_code, plaintext, plaintext_length,
-		    ciphertext, cb, aad, aad_length, tag, AES_BLOCK_SIZE, key, subkey);
+		    ciphertext, cb, aad, aad_length, tag, key, subkey);
 		if (rc)
 			return rc;
 	}
@@ -2181,7 +2491,7 @@ unsigned int ica_aes_gcm_last( unsigned char *icb,
 		if (rc)
 			return rc;
 
-		if (memcmp(tag, final_tag, final_tag_length))
+		if (CRYPTO_memcmp(tag, final_tag, final_tag_length))
 			return EFAULT;
 	}
 	return 0;
